@@ -36,11 +36,10 @@ import dkimpy_milter.config as config
 from dkimpy_milter.util import drop_privileges
 from dkimpy_milter.util import setExceptHook
 from dkimpy_milter.util import write_pid
-from dkimpy_milter.util import read_keyfile
-from dkimpy_milter.util import own_socketfile
+from dkimpy_milter.util import get_keys
 from dkimpy_milter.util import fold
 
-__version__ = "1.0.1"
+__version__ = "1.2.0"
 FWS = re.compile(r'\r?\n[ \t]+')
 
 
@@ -52,10 +51,9 @@ class dkimMilter(Milter.Base):
         self.id = Milter.uniqueID()
         # we don't want config used to change during a connection
         self.conf = milterconfig
-        self.privatersa = privateRSA
-        self.privateed25519 = privateEd25519
         self.fp = None
         self.fdomain = ''
+        self.iequals = None
 
     @Milter.noreply
     def connect(self, hostname, unused, hostaddr):
@@ -67,19 +65,19 @@ class dkimMilter(Milter.Base):
         if self.receiver is not None:
             self.receiver = self.receiver.strip()
         try:
-            self.AuthservID = milterconfig['AuthservID']
+            self.AuthservID = self.conf['AuthservID']
         except:
             self.AuthservID = self.receiver
         if hostaddr and len(hostaddr) > 0:
             ipaddr = hostaddr[0]
-            if milterconfig['IntHosts']:
-                if milterconfig['IntHosts'].match(ipaddr):
+            if self.conf['IntHosts']:
+                if self.conf['IntHosts'].match(ipaddr):
                     self.internal_connection = True
         else:
             ipaddr = ''
         self.connectip = ipaddr
-        if milterconfig.get('MacroList') and not self.internal_connection:
-            macrolist = milterconfig.get('MacroList')
+        if self.conf.get('MacroList') and not self.internal_connection:
+            macrolist = self.conf.get('MacroList')
             for macro in macrolist:
                 macroname = macro.split('|')[0]
                 macroname = '{' + macroname + '}'
@@ -87,8 +85,8 @@ class dkimMilter(Milter.Base):
                 if ((len(macro.split('|')) == 1 and macroresult) or macroresult
                         in macro.split('|')[1:]):
                     self.internal_connection = True
-        if milterconfig.get('MacroListVerify'):
-            macrolist = milterconfig.get('MacroListVerify')
+        if self.conf.get('MacroListVerify'):
+            macrolist = self.conf.get('MacroListVerify')
             for macro in macrolist:
                 macroname = macro.split('|')[0]
                 macroname = '{' + macroname + '}'
@@ -100,7 +98,7 @@ class dkimMilter(Milter.Base):
             connecttype = 'INTERNAL'
         else:
             connecttype = 'EXTERNAL'
-        if milterconfig.get('Syslog') and milterconfig.get('debugLevel') >= 1:
+        if self.conf.get('Syslog') and self.conf.get('debugLevel') >= 1:
             syslog.syslog("connect from {0} at {1} {2}"
                           .format(hostname, hostaddr, connecttype))
         return Milter.CONTINUE
@@ -110,7 +108,7 @@ class dkimMilter(Milter.Base):
     # of each message.
     @Milter.noreply
     def envfrom(self, f, *str):
-        if milterconfig.get('Syslog') and milterconfig.get('debugLevel') >= 2:
+        if self.conf.get('Syslog') and self.conf.get('debugLevel') >= 2:
             syslog.syslog("mail from: {0} {1}".format(f, str))
         self.fp = io.BytesIO()
         self.mailfrom = f
@@ -127,9 +125,14 @@ class dkimMilter(Milter.Base):
     @Milter.noreply
     def header(self, name, val):
         lname = name.lower()
+        if self.conf.get('Syslog') and self.conf.get('debugLevel') >= 4:
+            if lname == 'content-transfer-encoding':
+                syslog.syslog("content-transfer-encodeing: {0}".format(val))
+            if lname == 'content-type':
+                syslog.syslog("content-type: {0}".format(val))
         if lname == 'dkim-signature':
-            if (milterconfig.get('Syslog') and
-                    milterconfig.get('debugLevel') >= 1):
+            if (self.conf.get('Syslog') and
+                    self.conf.get('debugLevel') >= 1):
                 syslog.syslog("{0}: {1}".format(name, val))
             self.has_dkim += 1
         if lname == 'from':
@@ -138,8 +141,8 @@ class dkimMilter(Milter.Base):
                 self.fdomain = self.author.split('@')[1].lower()
             except IndexError as er:
                 pass # self.author was not a proper email address
-            if (milterconfig.get('Syslog') and
-                    milterconfig.get('debugLevel') >= 1):
+            if (self.conf.get('Syslog') and
+                    self.conf.get('debugLevel') >= 1):
                 syslog.syslog("{0}: {1}".format(name, val))
         elif lname == 'authentication-results':
             self.arheaders.append(val)
@@ -176,42 +179,126 @@ class dkimMilter(Milter.Base):
                       .parse_value(FWS.sub('', val)))
                 if ar.authserv_id == self.AuthservID:
                     self.chgheader('authentication-results', i, '')
-                    if (milterconfig.get('Syslog') and
-                            milterconfig.get('debugLevel') >= 1):
+                    if (self.conf.get('Syslog') and
+                            self.conf.get('debugLevel') >= 1):
                         syslog.syslog('REMOVE: {0}'.format(val))
             except:
                 # Don't error out on unparseable AR header fiels
                 pass
         # Check and/or sign DKIM
+        if (self.conf.get('Syslog') and self.conf.get('debugLevel') >= 4):
+            syslog.syslog('self.conf: {0}'.format(self.conf))
         self.fp.seek(0)
         txt = self.fp.read()
-        if milterconfig.get('Domain'):
-            domain = milterconfig.get('Domain')
-        else:
-            domain = ''
-        if milterconfig.get('SubDomains'):
-            self.fdomain = _get_parent_domain(self.fdomain, domain)
-        if ((self.fdomain in domain) and not milterconfig.get('Mode') == 'v'
+        self.get_identities_sign()
+        if (self.conf.get('Syslog') and self.conf.get('debugLevel') >= 3):
+            syslog.syslog('self.domain: {0}, self.fdomain: {1}, self.iequals: {2}'.format(self.domain, self.fdomain, self.iequals))
+        if ((self.fdomain in self.domain) and not self.conf.get('Mode') == 'v'
                 and not self.external_connection):
             self.sign_dkim(txt)
         if ((self.has_dkim) and (not self.internal_connection) and
-            (milterconfig.get('Mode') == 'v' or
-             milterconfig.get('Mode') == 'sv')):
+            (self.conf.get('Mode') == 'v' or
+             self.conf.get('Mode') == 'sv')):
             self.check_dkim(txt)
         if self.arresults:
             h = authres.AuthenticationResultsHeader(authserv_id=
                                                     self.AuthservID,
                                                     results=self.arresults)
             h = fold(codecs.encode(str(h), 'ascii'))
-            if (milterconfig.get('Syslog') and
-                    milterconfig.get('debugLevel') >= 2):
+            if (self.conf.get('Syslog') and
+                    self.conf.get('debugLevel') >= 2):
                 syslog.syslog(codecs.decode(h, 'ascii'))
             name, val = codecs.decode(h, 'ascii').split(': ', 1)
             self.addheader(name, val, 0)
         return Milter.CONTINUE
 
+    # get parent domain to be signed for if fdomain is a subdomain
+    def get_parent_domain(self, fdomain, domains):
+        for domain in domains:
+            rhs = '.'+domain
+            # compare right hand side of fdomain against .domain
+            if fdomain[-len(rhs):] == rhs:
+                # return parent domain on match
+                syslog.syslog('domain: {0}'.format(domain))
+                return domain
+        # or return the fdomain itself
+        return fdomain
+
+    def get_identities_sign(self):
+        """Determine d= and i= identiies for signature"""
+        self.domain = []
+        iequals = None
+        try:
+            self.privkeyRSA = self.conf.get('privateRSA')
+        except:
+            self.privkeyRSA = ''
+        try:
+            self.privkeyEd25519 = self.conf.get('privateEd25519')
+        except:
+            self.privkeyEd25519 = ''
+        try:
+            self.selectorRSA = self.conf.get('Selector')
+        except:
+            self.selectorRSA = ''
+        try:
+            self.selectorEd25519 = self.conf.get('SelectorEd25519')
+        except:
+            self.selectorEd25519 = ''
+        if not self.domain and self.conf.get('Domain'):
+            self.domain = self.conf.get('Domain')
+        if self.conf.get('SubDomains'):
+            self.fdomain = self.get_parent_domain(self.fdomain, self.domain)
+        if self.conf.get('SigningTable'):
+            match = False
+            for dictkey, dictvalues in self.conf.get('SigningTable').items():
+                if dictkey == '%':
+                    self.domain.append(self.fdomain)
+                    match = True
+                elif len(dictkey.split('*')) == 1:
+                    if dictkey == self.author:
+                        self.domain.append(self.fdomain)
+                        match = True
+                else:
+                    if len(dictkey.split('*')) == 2:
+                        if dictkey.split('*')[1] == self.author[-len(dictkey.split('*')[1]):]:
+                            self.domain.append(self.fdomain)
+                            match = True
+                    self.domain.append(self.fdomain)
+                try:
+                    if len(dictvalues) == 2 and match:
+                        if dictvalues[0] =='%':
+                            self.iequals = codecs.encode('@' + self.fdomain)
+                        elif dictvalues[0][1:] == self.fdomain or self.get_parent_domain(dictvalues[0][1:], self.domain) == self.fdomain:
+                            self.iequals = codecs.encode(dictvalues[0])
+                except IndexError:
+                    pass
+                if match:
+                    #TODO add KeyTable stuffs here.
+                    keytablekey = dictvalues[-1] # Last value in the SigningTable row.
+                    if self.conf.get('privateRSATable'):
+                        # Table data is a list of [ signing domain, selector, key ]
+                        keytabledata = self.conf.get('privateRSATable')[keytablekey]
+                        try:
+                            self.fdomain = keytabledata[0]
+                            self.selectorRSA = keytabledata[1]
+                            self.privkeyRSA = keytabledata[2]
+                        except:
+                            if (self.conf.get('Syslog')):
+                                    syslog.syslog('Error: Invalid KeyTable data {0}'.format(keytabledata))
+                    if self.conf.get('privateEd25519Table'):
+                        # Table data is a list of [ signing domain, selector, key ]
+                        keytabledata = self.conf.get('privateEd25519Table')[keytablekey]
+                        try:
+                            self.fdomain = keytabledata[0]
+                            self.selectorEd25519 = keytabledata[1]
+                            self.privkeyEd25519 = keytabledata[2]
+                        except:
+                            if (self.conf.get('Syslog')):
+                                    syslog.syslog('Error: Invalid KeyTable data {0}'.format(keytabledata))
+                    break
+
     def sign_dkim(self, txt):
-        canon = codecs.encode(milterconfig.get('Canonicalization'), 'ascii')
+        canon = codecs.encode(self.conf.get('Canonicalization'), 'ascii')
         canonicalize = []
         if len(canon.split(b'/')) == 2:
             canonicalize.append(canon.split(b'/')[0])
@@ -219,47 +306,52 @@ class dkimMilter(Milter.Base):
         else:
             canonicalize.append(canon)
             canonicalize.append(canon)
-            if (milterconfig.get('Syslog') and
-                    milterconfig.get('debugLevel') >= 1):
+            if (self.conf.get('Syslog') and
+                    self.conf.get('debugLevel') >= 1):
                 syslog.syslog('canonicalize: {0}'.format(canonicalize))
+        sign_headers = self.conf.get('SignHeaders')
+        if not sign_headers:
+            # None or empty. DKIM explicitly tests for None.
+            sign_headers = None
         try:
-            if privateRSA:
+            if self.privkeyRSA:
                 d = dkim.DKIM(txt)
-                h = d.sign(codecs.encode(milterconfig.get('Selector'), 'ascii'), codecs.encode(self.fdomain, 'ascii'),
-                           codecs.encode(privateRSA, 'ascii'),
-                           canonicalize=(canonicalize[0],
-                                         canonicalize[1]))
+                h = d.sign(codecs.encode(self.selectorRSA, 'ascii'), codecs.encode(self.fdomain, 'ascii'),
+                           codecs.encode(self.privkeyRSA, 'ascii'),
+                           canonicalize=(canonicalize[0], canonicalize[1]),
+                           identity=self.iequals, include_headers=sign_headers)
                 name, val = h.split(b': ', 1)
                 self.addheader(codecs.decode(name, 'ascii'), codecs.decode(val, 'ascii').strip().replace('\r\n', '\n'), 0)
-                if (milterconfig.get('Syslog') and
-                    (milterconfig.get('SyslogSuccess')
-                     or milterconfig.get('debugLevel') >= 1)):
+                if (self.conf.get('Syslog') and
+                    (self.conf.get('SyslogSuccess')
+                     or self.conf.get('debugLevel') >= 1)):
                     syslog.syslog('{0}: {1} DKIM signature added (s={2} '
                                   'd={3})'.format(self.getsymval('i'),
                                                   d.signature_fields.get(b'a').decode(),
                                                   d.signature_fields.get(b's').decode(),
                                                   d.domain.decode().lower()))
-            if privateEd25519:
+            if self.privkeyEd25519:
                 d = dkim.DKIM(txt)
-                h = d.sign(codecs.encode(milterconfig.get('SelectorEd25519'), 'ascii'), codecs.encode(self.fdomain, 'ascii'),
-                           privateEd25519, canonicalize=(canonicalize[0],
-                                                         canonicalize[1]),
+                h = d.sign(codecs.encode(self.selectorEd25519, 'ascii'), codecs.encode(self.fdomain, 'ascii'),
+                           self.privkeyEd25519,
+                           canonicalize=(canonicalize[0], canonicalize[1]),
+                           identity=self.iequals, include_headers=sign_headers,
                            signature_algorithm=b'ed25519-sha256')
                 name, val = h.split(b': ', 1)
                 self.addheader(codecs.decode(name, 'ascii'), codecs.decode(val, 'ascii').strip().replace('\r\n', '\n'), 0)
-                if (milterconfig.get('Syslog') and
-                    (milterconfig.get('SyslogSuccess')
-                     or milterconfig.get('debugLevel') >= 1)):
+                if (self.conf.get('Syslog') and
+                    (self.conf.get('SyslogSuccess')
+                     or self.conf.get('debugLevel') >= 1)):
                     syslog.syslog('{0}: {1} DKIM signature added (s={2} '
                                   'd={3})'.format(self.getsymval('i'),
                                                   d.signature_fields.get(b'a').decode(),
                                                   d.signature_fields.get(b's').decode(),
                                                   d.domain.decode().lower()))
         except dkim.DKIMException as x:
-            if milterconfig.get('Syslog'):
+            if self.conf.get('Syslog'):
                 syslog.syslog('DKIM: {0}'.format(x))
         except Exception as x:
-            if milterconfig.get('Syslog'):
+            if self.conf.get('Syslog'):
                 syslog.syslog("sign_dkim: {0}".format(x))
             raise
 
@@ -267,9 +359,9 @@ class dkimMilter(Milter.Base):
         res = False
         self.header_a = None
         for y in range(self.has_dkim):  # Verify _ALL_ the signatures
-            d = dkim.DKIM(txt)
+            d = dkim.DKIM(txt, minkey=self.conf.get('MinimumKeyBits'), timeout=self.conf.get('DNSTimeout'))
             try:
-                dnsoverride = milterconfig.get('DNSOverride')
+                dnsoverride = self.conf.get('DNSOverride')
                 if isinstance(dnsoverride, str):
                     syslog.syslog("DNSOverride: {0}".format(dnsoverride))
                     res = d.verify(idx=y, dnsfunc=lambda _x: dnsoverride)
@@ -288,12 +380,12 @@ class dkimMilter(Milter.Base):
                                          .format(d.keysize, algo))
             except dkim.DKIMException as x:
                 self.dkim_comment = str(x)
-                if milterconfig.get('Syslog'):
+                if self.conf.get('Syslog'):
                     syslog.syslog('DKIM: {0}'.format(x))
             except Exception as x:
                 self.dkim_comment = str(x)
-                if milterconfig.get('Syslog'):
-                    syslog.syslog("check_dkim: {0}".format(x))
+                if self.conf.get('Syslog'):
+                    syslog.syslog("check_dkim: Internal program fault while verifying: {0}".format(x))
             try:
                 # i= is optional and dkimpy is fine if it's not provided
                 self.header_i = codecs.decode(d.signature_fields.get(b'i'), 'ascii')
@@ -304,15 +396,15 @@ class dkimMilter(Milter.Base):
                 self.header_a = codecs.decode(d.signature_fields.get(b'a'), 'ascii')
             except Exception as x:
                 self.dkim_comment = str(x)
-                if milterconfig.get('Syslog'):
-                    syslog.syslog("check_dkim: {0}".format(x))
+                if self.conf.get('Syslog'):
+                    syslog.syslog("check_dkim: Internal program fault extracting header a or d: {0}".format(x))
                 self.header_d = None
             if not self.header_a:
                 self.header_a = 'rsa-sha256'
             if res:
-                if (milterconfig.get('Syslog') and
-                        (milterconfig.get('SyslogSuccess') or
-                         milterconfig.get('debugLevel') >= 1)):
+                if (self.conf.get('Syslog') and
+                        (self.conf.get('SyslogSuccess') or
+                         self.conf.get('debugLevel') >= 1)):
                     syslog.syslog('{0}: {1} DKIM signature verified (s={2} '
                                   'd={3})'.format(self.getsymval('i'),
                                                   d.signature_fields.get(b'a').decode(),
@@ -320,15 +412,16 @@ class dkimMilter(Milter.Base):
                                                   d.domain.decode().lower()))
                 self.dkim_domain = d.domain.lower()
             else:
-                if milterconfig.get('DiagnosticDirectory'):
+                if self.conf.get('DiagnosticDirectory'):
+                    tempfile.tempdir = self.conf.get('DiagnosticDirectory')
                     fd, fname = tempfile.mkstemp(".dkim")
                     with os.fdopen(fd, "w+b") as fp:
                         fp.write(txt)
-                    if milterconfig.get('Syslog'):
+                    if self.conf.get('Syslog'):
                         syslog.syslog('DKIM: Fail (saved as {0})'
                                       .format(fname))
                 else:
-                    if milterconfig.get('Syslog'):
+                    if self.conf.get('Syslog'):
                         if d.domain:
                             syslog.syslog('DKIM: Fail ({0})'
                                           .format(d.domain.lower()))
@@ -351,41 +444,28 @@ class dkimMilter(Milter.Base):
             self.header_a = None
         return
 
-# get parent domain to be signed for if fdomain is a subdomain
-def _get_parent_domain(fdomain, domains):
-    for domain in domains:
-        rhs = '.'+domain
-        # compare right hand side of fdomain against .domain
-        if fdomain[-len(rhs):] == rhs:
-            # return parent domain on match
-            return domain
-    # or return the fdomain itself
-    return fdomain
-
 def main():
     # Ugh, but there's no easy way around this.
     global milterconfig
-    global privateRSA
-    global privateEd25519
-    privateRSA = False
-    privateEd25519 = False
     configFile = '/usr/local/etc/dkimpy-milter.conf'
     if len(sys.argv) > 1:
-        if sys.argv[1] in ('-?', '--help', '-h'):
-            print('usage: dkimpy-milter [<configfilename>]')
+        if (sys.argv[1] in ('-?', '--help', '-h')) or len(sys.argv) == 3 or \
+               (len(sys.argv) == 4 and sys.argv[2] != '-P'):
+            print('usage: dkimpy-milter [<configfilename> [-P <pidfile>]]')
             sys.exit(1)
         configFile = sys.argv[1]
     milterconfig = config._processConfigFile(filename=configFile)
+    if len(sys.argv) == 4:
+        if sys.argv[2] == '-P':
+            # Command line PID file argument overrides config file
+            milterconfig['PidFile'] = sys.argv[3]
     if milterconfig.get('Syslog'):
         facility = eval("syslog.LOG_{0}"
                         .format(milterconfig.get('SyslogFacility').upper()))
         syslog.openlog(os.path.basename(sys.argv[0]), syslog.LOG_PID, facility)
         setExceptHook()
     pid = write_pid(milterconfig)
-    if milterconfig.get('KeyFile'):
-        privateRSA = read_keyfile(milterconfig, 'RSA')
-    if milterconfig.get('KeyFileEd25519'):
-        privateEd25519 = read_keyfile(milterconfig, 'Ed25519')
+    milterconfig = get_keys(milterconfig)
     Milter.factory = dkimMilter
     Milter.set_flags(Milter.CHGHDRS + Milter.ADDHDRS)
     miltername = 'dkimpy-filter'
@@ -401,12 +481,11 @@ def main():
                     socketname = 'fd:3'
         if socketname is None:
             socketname = 'local:/var/run/dkimpy-milter/dkimpy-milter.sock'
-    own_socketfile(milterconfig, socketname)
-    drop_privileges(milterconfig)
     sys.stdout.flush()
     if milterconfig.get('Syslog'):
         syslog.syslog('dkimpy-milter starting:{0} user:{1}'
                       .format(pid, milterconfig.get('UserID')))
+    drop_privileges(milterconfig)
     Milter.runmilter(miltername, socketname, 240)
 
 if __name__ == "__main__":
